import os
import os.path
import sys
import time
import shutil

from ctypesgen.ctypedescs import CtypesBitfield, CtypesStruct
from ctypesgen.expressions import ExpressionNode
from ctypesgen.messages import error_message, status_message


THIS_DIR = os.path.dirname(__file__)
CTYPESGEN_DIR = os.path.join(THIS_DIR, os.path.pardir)
PREAMBLE_PATH = os.path.join(THIS_DIR, "preamble.py")
DEFAULTHEADER_PATH = os.path.join(THIS_DIR, "defaultheader.py")
LIBRARYLOADER_PATH = os.path.join(CTYPESGEN_DIR, "libraryloader.py")


class WrapperPrinter:
    def __init__(self, outpath, options, data):
        status_message("Writing to %s." % (outpath or "stdout"))

        self.file = open(outpath, "w") if outpath else sys.stdout
        self.options = options

        if self.options.strip_build_path and self.options.strip_build_path[-1] != os.path.sep:
            self.options.strip_build_path += os.path.sep

        if not self.options.embed_preamble and outpath:
            self._copy_preamble_loader_files(outpath)

        self.print_header()
        self.file.write("\n")

        self.print_preamble()
        self.file.write("\n")

        self.print_loader()
        self.file.write("\n")

        self.print_group(self.options.libraries, "libraries", self.print_library)
        self.print_group(self.options.modules, "modules", self.print_module)

        method_table = {
            "function": self.print_function,
            "macro": self.print_macro,
            "struct": self.print_struct,
            "struct-body": self.print_struct_members,
            "typedef": self.print_typedef,
            "variable": self.print_variable,
            "enum": self.print_enum,
            "constant": self.print_constant,
            "undef": self.print_undef,
        }

        for kind, desc in data.output_order:
            if desc.included:
                method_table[kind](desc)
                self.file.write("\n")

        self.print_group(self.options.inserted_files, "inserted files", self.insert_file)
        self.strip_prefixes()

    def __del__(self):
        self.file.close()

    def print_group(self, list, name, function):
        if list:
            self.file.write("# Begin %s\n" % name)
            for obj in list:
                function(obj)
            self.file.write("\n")
            self.file.write("# %d %s\n" % (len(list), name))
            self.file.write("# End %s\n" % name)
        else:
            self.file.write("# No %s\n" % name)
        self.file.write("\n")

    def srcinfo(self, src):
        if src is None:
            self.file.write("\n")
        else:
            filename, lineno = src
            if filename in ("<built-in>", "<command line>"):
                self.file.write("# %s\n" % filename)
            else:
                if self.options.strip_build_path and filename.startswith(
                    self.options.strip_build_path
                ):
                    filename = filename[len(self.options.strip_build_path) :]
                self.file.write("# %s: %s\n" % (filename, lineno))

    def template_subs(self):
        template_subs = {
            "date": time.ctime(),
            "argv": " ".join([x for x in sys.argv if not x.startswith("--strip-build-path")]),
            "name": os.path.basename(self.options.headers[0]),
        }

        for opt, value in self.options.__dict__.items():
            if type(value) == str:
                template_subs[opt] = value
            elif isinstance(value, (list, tuple)):
                template_subs[opt] = (os.path.sep).join(value)
            else:
                template_subs[opt] = repr(value)

        return template_subs

    def print_header(self):
        template_file = None

        if self.options.header_template:
            path = self.options.header_template
            try:
                template_file = open(path, "r")
            except IOError:
                error_message(
                    'Cannot load header template from file "%s" '
                    " - using default template." % path,
                    cls="missing-file",
                )

        if not template_file:
            template_file = open(DEFAULTHEADER_PATH, "r")

        template_subs = self.template_subs()
        self.file.write(template_file.read() % template_subs)

        template_file.close()

    def print_preamble(self):
        self.file.write("# Begin preamble for Python\n\n")
        if self.options.embed_preamble:
            with open(PREAMBLE_PATH, "r") as preamble_file:
                preamble_file_content = preamble_file.read()
                filecontent = preamble_file_content.replace("# ~POINTER~", "")
                self.file.write(filecontent)
        else:
            self.file.write("from .ctypes_preamble import *\n")
            self.file.write("from .ctypes_preamble import _variadic_function\n")

        self.file.write("\n# End preamble\n")

    def _copy_preamble_loader_files(self, path):
        if os.path.isfile(path):
            abspath = os.path.abspath(path)
            dst = os.path.dirname(abspath)
        else:
            error_message(
                "Cannot copy preamble and loader files",
                cls="missing-file",
            )
            return

        c_preamblefile = f"{dst}/ctypes_preamble.py"
        if os.path.isfile(c_preamblefile):
            return

        pointer = """def POINTER(obj):
    p = ctypes.POINTER(obj)

    # Convert None to a real NULL pointer to work around bugs
    # in how ctypes handles None on 64-bit platforms
    if not isinstance(p.from_param, classmethod):

        def from_param(cls, x):
            if x is None:
                return cls()
            else:
                return x

        p.from_param = classmethod(from_param)

    return p

"""

        with open(PREAMBLE_PATH) as preamble_file:
            preamble_file_content = preamble_file.read()
            filecontent = preamble_file_content.replace("# ~POINTER~", pointer)

        with open(c_preamblefile, "w") as f:
            f.write(filecontent)

        shutil.copy(LIBRARYLOADER_PATH, f"{dst}")
        os.rename(f"{dst}/libraryloader.py", f"{dst}/ctypes_loader.py")

    def print_loader(self):
        self.file.write("_libs = {}\n")
        self.file.write("_libdirs = %s\n\n" % self.options.compile_libdirs)
        self.file.write("# Begin loader\n\n")
        if self.options.embed_preamble:
            with open(LIBRARYLOADER_PATH, "r") as loader_file:
                self.file.write(loader_file.read())
        else:
            self.file.write("from .ctypes_loader import *\n")
        self.file.write("\n# End loader\n\n")
        self.file.write(
            "add_library_search_dirs([%s])"
            % ", ".join([repr(d) for d in self.options.runtime_libdirs])
        )
        self.file.write("\n")

    def print_library(self, library):
        self.file.write('_libs["%s"] = load_library("%s")\n' % (library, library))

    def print_module(self, module):
        self.file.write("from %s import *\n" % module)

    def print_constant(self, constant):
        self.file.write("%s = %s" % (constant.name, constant.value.py_string(False)))
        self.srcinfo(constant.src)

    def print_undef(self, undef):
        self.srcinfo(undef.src)
        self.file.write(
            "# #undef {macro}\n"
            "try:\n"
            "    del {macro}\n"
            "except NameError:\n"
            "    pass\n".format(macro=undef.macro.py_string(False))
        )

    def print_typedef(self, typedef):
        self.file.write("%s = %s" % (typedef.name, typedef.ctype.py_string()))
        self.srcinfo(typedef.src)

    def print_struct(self, struct):
        self.srcinfo(struct.src)
        base = {"union": "Union", "struct": "Structure"}[struct.variety]
        self.file.write("class %s_%s(%s):\n" "    pass\n" % (struct.variety, struct.tag, base))

    def print_struct_members(self, struct):
        if struct.opaque:
            return

        # is this supposed to be packed?
        if struct.attrib.get("packed", False):
            aligned = struct.attrib.get("aligned", [1])
            assert len(aligned) == 1, "cgrammar gave more than one arg for aligned attribute"
            aligned = aligned[0]
            if isinstance(aligned, ExpressionNode):
                # TODO: for non-constant expression nodes, this will fail:
                aligned = aligned.evaluate(None)
            self.file.write("{}_{}._pack_ = {}\n".format(struct.variety, struct.tag, aligned))

        # handle unnamed fields.
        unnamed_fields = []
        names = set([x[0] for x in struct.members])
        anon_prefix = "unnamed_"
        n = 1
        for mi in range(len(struct.members)):
            mem = list(struct.members[mi])
            if mem[0] is None:
                while True:
                    name = "%s%i" % (anon_prefix, n)
                    n += 1
                    if name not in names:
                        break
                mem[0] = name
                names.add(name)
                if type(mem[1]) is CtypesStruct:
                    unnamed_fields.append(name)
                struct.members[mi] = mem

        self.file.write("%s_%s.__slots__ = [\n" % (struct.variety, struct.tag))
        for name, ctype in struct.members:
            self.file.write("    '%s',\n" % name)
        self.file.write("]\n")

        if len(unnamed_fields) > 0:
            self.file.write("%s_%s._anonymous_ = [\n" % (struct.variety, struct.tag))
            for name in unnamed_fields:
                self.file.write("    '%s',\n" % name)
            self.file.write("]\n")

        self.file.write("%s_%s._fields_ = [\n" % (struct.variety, struct.tag))
        for name, ctype in struct.members:
            if isinstance(ctype, CtypesBitfield):
                self.file.write(
                    "    ('%s', %s, %s),\n"
                    % (name, ctype.py_string(), ctype.bitfield.py_string(False))
                )
            else:
                self.file.write("    ('%s', %s),\n" % (name, ctype.py_string()))
        self.file.write("]\n")

    def print_enum(self, enum):
        self.file.write("enum_%s = c_int" % enum.tag)
        self.srcinfo(enum.src)
        # Values of enumerator are output as constants.

    def print_function(self, function):
        if function.variadic:
            self.print_variadic_function(function)
        else:
            self.print_fixed_function(function)

    def print_fixed_function(self, function):
        self.srcinfo(function.src)

        CC = "stdcall" if function.attrib.get("stdcall", False) else "cdecl"

        # If we know what library the function lives in, look there.
        # Otherwise, check all the libraries.
        if function.source_library:
            self.file.write(
                'if _libs["{L}"].has("{CN}", "{CC}"):\n'
                '    {PN} = _libs["{L}"].get("{CN}", "{CC}")\n'.format(
                    L=function.source_library, CN=function.c_name(), PN=function.py_name(), CC=CC
                )
            )
        else:
            self.file.write(
                "for _lib in _libs.values():\n"
                '    if not _lib.has("{CN}", "{CC}"):\n'
                "        continue\n"
                '    {PN} = _lib.get("{CN}", "{CC}")\n'.format(
                    CN=function.c_name(), PN=function.py_name(), CC=CC
                )
            )

        # Argument types
        self.file.write(
            "    %s.argtypes = [%s]\n"
            % (function.py_name(), ", ".join([a.py_string() for a in function.argtypes]))
        )

        # Return value
        if function.restype.py_string() == "String":
            self.file.write(
                "    if sizeof(c_int) == sizeof(c_void_p):\n"
                "        {PN}.restype = ReturnString\n"
                "    else:\n"
                "        {PN}.restype = {RT}\n"
                "        {PN}.errcheck = ReturnString\n".format(
                    PN=function.py_name(), RT=function.restype.py_string()
                )
            )
        else:
            self.file.write(
                "    %s.restype = %s\n" % (function.py_name(), function.restype.py_string())
            )
            if function.errcheck:
                self.file.write(
                    "    %s.errcheck = %s\n" % (function.py_name(), function.errcheck.py_string())
                )

        if not function.source_library:
            self.file.write("    break\n")

    def print_variadic_function(self, function):
        CC = "stdcall" if function.attrib.get("stdcall", False) else "cdecl"

        self.srcinfo(function.src)
        if function.source_library:
            self.file.write(
                'if _libs["{L}"].has("{CN}", "{CC}"):\n'
                '    _func = _libs["{L}"].get("{CN}", "{CC}")\n'
                "    _restype = {RT}\n"
                "    _errcheck = {E}\n"
                "    _argtypes = [{t0}]\n"
                "    {PN} = _variadic_function(_func,_restype,_argtypes,_errcheck)\n".format(
                    L=function.source_library,
                    CN=function.c_name(),
                    RT=function.restype.py_string(),
                    E=function.errcheck.py_string(),
                    t0=", ".join([a.py_string() for a in function.argtypes]),
                    PN=function.py_name(),
                    CC=CC,
                )
            )
        else:
            self.file.write(
                "for _lib in _libs.values():\n"
                '    if _lib.has("{CN}", "{CC}"):\n'
                '        _func = _lib.get("{CN}", "{CC}")\n'
                "        _restype = {RT}\n"
                "        _errcheck = {E}\n"
                "        _argtypes = [{t0}]\n"
                "        {PN} = _variadic_function(_func,_restype,_argtypes,_errcheck)\n".format(
                    CN=function.c_name(),
                    RT=function.restype.py_string(),
                    E=function.errcheck.py_string(),
                    t0=", ".join([a.py_string() for a in function.argtypes]),
                    PN=function.py_name(),
                    CC=CC,
                )
            )

    def print_variable(self, variable):
        self.srcinfo(variable.src)
        if variable.source_library:
            self.file.write(
                "try:\n"
                '    {PN} = ({PS}).in_dll(_libs["{L}"], "{CN}")\n'
                "except:\n"
                "    pass\n".format(
                    PN=variable.py_name(),
                    PS=variable.ctype.py_string(),
                    L=variable.source_library,
                    CN=variable.c_name(),
                )
            )
        else:
            self.file.write(
                "for _lib in _libs.values():\n"
                "    try:\n"
                '        {PN} = ({PS}).in_dll(_lib, "{CN}")\n'
                "        break\n"
                "    except:\n"
                "        pass\n".format(
                    PN=variable.py_name(), PS=variable.ctype.py_string(), CN=variable.c_name()
                )
            )

    def print_macro(self, macro):
        if macro.params:
            self.print_func_macro(macro)
        else:
            self.print_simple_macro(macro)

    def print_simple_macro(self, macro):
        # The macro translator makes heroic efforts but it occasionally fails.
        # We want to contain the failures as much as possible.
        # Hence the try statement.
        self.srcinfo(macro.src)
        self.file.write(
            "try:\n"
            "    {MN} = {ME}\n"
            "except:\n"
            "    pass\n".format(MN=macro.name, ME=macro.expr.py_string(True))
        )

    def print_func_macro(self, macro):
        self.srcinfo(macro.src)
        self.file.write(
            "def {MN}({MP}):\n"
            "    return {ME}\n".format(
                MN=macro.name, MP=", ".join(macro.params), ME=macro.expr.py_string(True)
            )
        )

    def strip_prefixes(self):
        if not self.options.strip_prefixes:
            self.file.write("# No prefix-stripping\n\n")
            return

        self.file.write(
            "# Begin prefix-stripping\n"
            "\n"
            "# Strip prefixes from all symbols following regular expression:\n"
            "# {expr}\n"
            "\n"
            "import re as __re_module\n"
            "\n"
            "__strip_expr = __re_module.compile('{expr}')\n"
            "for __k, __v in globals().copy().items():\n"
            "    __m = __strip_expr.match(__k)\n"
            "    if __m:\n"
            "        globals()[__k[__m.end():]] = __v\n"
            "        # remove symbol with prefix(?)\n"
            "        # globals().pop(__k)\n"
            "del __re_module, __k, __v, __m, __strip_expr\n"
            "\n"
            "# End prefix-stripping\n"
            "\n".format(expr="({})".format("|".join(self.options.strip_prefixes)))
        )

    def insert_file(self, filename):
        try:
            inserted_file = open(filename, "r")
        except IOError:
            error_message('Cannot open file "%s". Skipped it.' % filename, cls="missing-file")

        self.file.write(
            '# Begin "{filename}"\n'
            "\n{file}\n"
            '# End "{filename}"\n'.format(filename=filename, file=inserted_file.read())
        )

        inserted_file.close()



"""
DataCollectingParser subclasses ctypesparser.CtypesParser and builds Description
objects from the CtypesType objects and other information from CtypesParser.
After parsing is complete, a DescriptionCollection object can be retrieved by
calling DataCollectingParser.data().
"""

import os
from tempfile import mkstemp

from ctypesgen.ctypedescs import CtypesEnum, CtypesType, CtypesTypeVisitor
from ctypesgen.descriptions import (
    ConstantDescription,
    DescriptionCollection,
    EnumDescription,
    FunctionDescription,
    MacroDescription,
    StructDescription,
    TypedefDescription,
    UndefDescription,
    VariableDescription,
)
from ctypesgen.expressions import ConstantExpressionNode
from ctypesgen.messages import error_message, status_message
from ctypesgen.parser import ctypesparser


class DataCollectingParser(ctypesparser.CtypesParser, CtypesTypeVisitor):
    """Main class for the Parser component. Steps for use:
    p=DataCollectingParser(names_of_header_files,options)
    p.parse()
    data=p.data() #A dictionary of constants, enums, structs, functions, etc.
    """

    def __init__(self, headers, options):
        super(DataCollectingParser, self).__init__(options)
        self.headers = headers
        self.options = options

        self.constants = []
        self.typedefs = []
        self.structs = []
        self.enums = []
        self.functions = []
        self.variables = []
        self.macros = []

        self.all = []
        self.output_order = []

        # NULL is a useful macro to have defined
        null = ConstantExpressionNode(None)
        nullmacro = ConstantDescription("NULL", null, ("<built-in>", 1))
        self.constants.append(nullmacro)
        self.all.append(nullmacro)
        self.output_order.append(("constant", nullmacro))

        # A list of tuples describing macros; saved to be processed after
        # everything else has been parsed
        self.saved_macros = []
        # A set of structs that are already known
        self.already_seen_structs = set()
        # A dict of structs that have only been seen in opaque form
        self.already_seen_opaque_structs = {}
        # A set of enums that are already known
        self.already_seen_enums = set()
        # A dict of enums that have only been seen in opaque form
        self.already_seen_opaque_enums = {}

    def parse(self):
        fd, fname = mkstemp(suffix=".h")
        with os.fdopen(fd, "w") as f:
            for header in self.options.other_headers:
                f.write("#include <%s>\n" % header)
            for header in self.headers:
                f.write('#include "%s"\n' % os.path.abspath(header))
            f.flush()
        try:
            super(DataCollectingParser, self).parse(fname, self.options.debug_level)
        finally:
            os.unlink(fname)

        for name, params, expr, (filename, lineno) in self.saved_macros:
            self.handle_macro(name, params, expr, filename, lineno)

    def handle_define_constant(self, name, expr, filename, lineno):
        # Called by CParser
        # Save to handle later
        self.saved_macros.append((name, None, expr, (filename, lineno)))

    def handle_define_unparseable(self, name, params, value, filename, lineno):
        # Called by CParser
        if params:
            original_string = "#define %s(%s) %s" % (name, ",".join(params), " ".join(value))
        else:
            original_string = "#define %s %s" % (name, " ".join(value))
        macro = MacroDescription(name, params, None, src=(filename, lineno))
        macro.error('Could not parse macro "%s"' % original_string, cls="macro")
        macro.original_string = original_string
        self.macros.append(macro)
        self.all.append(macro)
        self.output_order.append(("macro", macro))

    def handle_define_macro(self, name, params, expr, filename, lineno):
        # Called by CParser
        # Save to handle later
        self.saved_macros.append((name, params, expr, (filename, lineno)))

    def handle_undefine(self, macro, filename, lineno):
        # save to handle later to get order correct
        self.saved_macros.append(("#undef", None, macro, (filename, lineno)))

    def handle_ctypes_typedef(self, name, ctype, filename, lineno):
        # Called by CtypesParser
        ctype.visit(self)

        typedef = TypedefDescription(name, ctype, src=(filename, repr(lineno)))

        self.typedefs.append(typedef)
        self.all.append(typedef)
        self.output_order.append(("typedef", typedef))

    def handle_ctypes_new_type(self, ctype, filename, lineno):
        # Called by CtypesParser
        if isinstance(ctype, CtypesEnum):
            self.handle_enum(ctype, filename, lineno)
        else:
            self.handle_struct(ctype, filename, lineno)

    def handle_ctypes_function(
        self, name, restype, argtypes, errcheck, variadic, attrib, filename, lineno
    ):
        # Called by CtypesParser
        restype.visit(self)
        for argtype in argtypes:
            argtype.visit(self)

        function = FunctionDescription(
            name, restype, argtypes, errcheck, variadic, attrib, src=(filename, repr(lineno))
        )

        self.functions.append(function)
        self.all.append(function)
        self.output_order.append(("function", function))

    def handle_ctypes_variable(self, name, ctype, filename, lineno):
        # Called by CtypesParser
        ctype.visit(self)

        variable = VariableDescription(name, ctype, src=(filename, repr(lineno)))

        self.variables.append(variable)
        self.all.append(variable)
        self.output_order.append(("variable", variable))

    def handle_struct(self, ctypestruct, filename, lineno):
        # Called from within DataCollectingParser

        # When we find an opaque struct, we make a StructDescription for it
        # and record it in self.already_seen_opaque_structs. If we later
        # find a transparent struct with the same tag, we fill in the
        # opaque struct with the information from the transparent struct and
        # move the opaque struct to the end of the struct list.

        name = "%s %s" % (ctypestruct.variety, ctypestruct.tag)

        if name in self.already_seen_structs:
            return

        if ctypestruct.opaque:
            if name not in self.already_seen_opaque_structs:
                struct = StructDescription(
                    ctypestruct.tag,
                    ctypestruct.attrib,
                    ctypestruct.variety,
                    None,  # No members
                    True,  # Opaque
                    ctypestruct,
                    src=(filename, str(lineno)),
                )

                self.already_seen_opaque_structs[name] = struct
                self.structs.append(struct)
                self.all.append(struct)
                self.output_order.append(("struct", struct))

        else:
            for (membername, ctype) in ctypestruct.members:
                ctype.visit(self)

            if name in self.already_seen_opaque_structs:
                # Fill in older version
                struct = self.already_seen_opaque_structs[name]
                struct.opaque = False
                struct.members = ctypestruct.members
                struct.ctype = ctypestruct
                struct.src = ctypestruct.src

                self.output_order.append(("struct-body", struct))

                del self.already_seen_opaque_structs[name]

            else:
                struct = StructDescription(
                    ctypestruct.tag,
                    ctypestruct.attrib,
                    ctypestruct.variety,
                    ctypestruct.members,
                    False,  # Not opaque
                    src=(filename, str(lineno)),
                    ctype=ctypestruct,
                )
                self.structs.append(struct)
                self.all.append(struct)
                self.output_order.append(("struct", struct))
                self.output_order.append(("struct-body", struct))

            self.already_seen_structs.add(name)

    def handle_enum(self, ctypeenum, filename, lineno):
        # Called from within DataCollectingParser.

        # Process for handling opaque enums is the same as process for opaque
        # structs. See handle_struct() for more details.

        tag = ctypeenum.tag
        if tag in self.already_seen_enums:
            return

        if ctypeenum.opaque:
            if tag not in self.already_seen_opaque_enums:
                enum = EnumDescription(ctypeenum.tag, None, ctypeenum, src=(filename, str(lineno)))
                enum.opaque = True

                self.already_seen_opaque_enums[tag] = enum
                self.enums.append(enum)
                self.all.append(enum)
                self.output_order.append(("enum", enum))

        else:
            if tag in self.already_seen_opaque_enums:
                # Fill in older opaque version
                enum = self.already_seen_opaque_enums[tag]
                enum.opaque = False
                enum.ctype = ctypeenum
                enum.src = ctypeenum.src
                enum.members = ctypeenum.enumerators

                del self.already_seen_opaque_enums[tag]

            else:
                enum = EnumDescription(
                    ctypeenum.tag,
                    ctypeenum.enumerators,
                    src=(filename, str(lineno)),
                    ctype=ctypeenum,
                )
                enum.opaque = False

                self.enums.append(enum)
                self.all.append(enum)
                self.output_order.append(("enum", enum))

            self.already_seen_enums.add(tag)

            for (enumname, expr) in ctypeenum.enumerators:
                constant = ConstantDescription(enumname, expr, src=(filename, lineno))

                self.constants.append(constant)
                self.all.append(constant)
                self.output_order.append(("constant", constant))

    def handle_macro(self, name, params, expr, filename, lineno):
        # Called from within DataCollectingParser
        src = (filename, lineno)

        if expr is None:
            expr = ConstantExpressionNode(True)
            constant = ConstantDescription(name, expr, src)
            self.constants.append(constant)
            self.all.append(constant)
            return

        expr.visit(self)

        if isinstance(expr, CtypesType):
            if params:
                macro = MacroDescription(name, "", src)
                macro.error(
                    "%s has parameters but evaluates to a type. "
                    "Ctypesgen does not support it." % macro.casual_name(),
                    cls="macro",
                )
                self.macros.append(macro)
                self.all.append(macro)
                self.output_order.append(("macro", macro))

            else:
                typedef = TypedefDescription(name, expr, src)
                self.typedefs.append(typedef)
                self.all.append(typedef)
                self.output_order.append(("typedef", typedef))

        elif name == "#undef":
            undef = UndefDescription(expr, src)
            self.all.append(undef)
            self.output_order.append(("undef", undef))
        else:
            macro = MacroDescription(name, params, expr, src)
            self.macros.append(macro)
            self.all.append(macro)
            self.output_order.append(("macro", macro))

        # Macros could possibly contain things like __FILE__, __LINE__, etc...
        # This could be supported, but it would be a lot of work. It would
        # probably also bloat the Preamble considerably.

    def handle_error(self, message, filename, lineno):
        # Called by CParser
        error_message("%s:%d: %s" % (filename, lineno, message), cls="cparser")

    def handle_pp_error(self, message):
        # Called by PreprocessorParser
        error_message("%s: %s" % (self.options.cpp, message), cls="cparser")

    def handle_status(self, message):
        # Called by CParser
        status_message(message)

    def visit_struct(self, struct):
        self.handle_struct(struct, struct.src[0], struct.src[1])

    def visit_enum(self, enum):
        self.handle_enum(enum, enum.src[0], enum.src[1])

    def data(self):
        return DescriptionCollection(
            self.constants,
            self.typedefs,
            self.structs,
            self.enums,
            self.functions,
            self.variables,
            self.macros,
            self.all,
            self.output_order,
        )



"""
The dependencies module determines which descriptions depend on which other
descriptions.
"""

from ctypesgen.descriptions import MacroDescription, UndefDescription
from ctypesgen.ctypedescs import visit_type_and_collect_info


def find_dependencies(data, opts):
    """Visit each description in `data` and figure out which other descriptions
    it depends on, putting the results in desc.requirements. Also find errors in
    ctypedecls or expressions attached to the description and transfer them to the
    description."""

    struct_names = {}
    enum_names = {}
    typedef_names = {}
    ident_names = {}

    # Start the lookup tables with names from imported modules

    for name in opts.other_known_names:
        typedef_names[name] = None
        ident_names[name] = None
        if name.startswith("struct_") or name.startswith("enum_"):
            variety = name.split("_")[0]
            tag = "_".join(name.split("_")[1:])
            struct_names[(variety, tag)] = None
        if name.startswith("enum_"):
            enum_names[name] = None

    def depend(desc, nametable, name):
        """Try to add `name` as a requirement for `desc`, looking `name` up in
        `nametable`. Returns True if found."""

        if name in nametable:
            requirement = nametable[name]
            if requirement:
                desc.add_requirements([requirement])
            return True
        else:
            return False

    def co_depend(desc, nametable, name):
        """
        Try to add `name` as a requirement for `desc`, looking `name` up in
        `nametable`.  Also try to add desc as a requirement for `name`.

        Returns Description of `name` if found.
        """

        requirement = nametable.get(name, None)
        if requirement is None:
            return

        desc.add_requirements([requirement])
        requirement.add_requirements([desc])
        return requirement

    def find_dependencies_for(desc, kind):
        """Find all the descriptions that `desc` depends on and add them as
        dependencies for `desc`. Also collect error messages regarding `desc` and
        convert unlocateable descriptions into error messages."""

        if kind == "constant":
            roots = [desc.value]
        elif kind == "struct":
            roots = []
        elif kind == "struct-body":
            roots = [desc.ctype]
        elif kind == "enum":
            roots = []
        elif kind == "typedef":
            roots = [desc.ctype]
        elif kind == "function":
            roots = desc.argtypes + [desc.restype]
        elif kind == "variable":
            roots = [desc.ctype]
        elif kind == "macro":
            if desc.expr:
                roots = [desc.expr]
            else:
                roots = []
        elif kind == "undef":
            roots = [desc.macro]

        cstructs, cenums, ctypedefs, errors, identifiers = [], [], [], [], []

        for root in roots:
            s, e, t, errs, i = visit_type_and_collect_info(root)
            cstructs.extend(s)
            cenums.extend(e)
            ctypedefs.extend(t)
            errors.extend(errs)
            identifiers.extend(i)

        unresolvables = []

        for cstruct in cstructs:
            if kind == "struct" and desc.variety == cstruct.variety and desc.tag == cstruct.tag:
                continue
            if not depend(desc, struct_names, (cstruct.variety, cstruct.tag)):
                unresolvables.append('%s "%s"' % (cstruct.variety, cstruct.tag))

        for cenum in cenums:
            if kind == "enum" and desc.tag == cenum.tag:
                continue
            if not depend(desc, enum_names, cenum.tag):
                unresolvables.append('enum "%s"' % cenum.tag)

        for ctypedef in ctypedefs:
            if not depend(desc, typedef_names, ctypedef):
                unresolvables.append('typedef "%s"' % ctypedef)

        for ident in identifiers:
            if isinstance(desc, MacroDescription) and desc.params and ident in desc.params:
                continue

            elif opts.include_undefs and isinstance(desc, UndefDescription):
                macro_desc = None
                if ident == desc.macro.name:
                    macro_desc = co_depend(desc, ident_names, ident)
                if macro_desc is None or not isinstance(macro_desc, MacroDescription):
                    unresolvables.append('identifier "%s"' % ident)

            elif not depend(desc, ident_names, ident):
                unresolvables.append('identifier "%s"' % ident)

        for u in unresolvables:
            errors.append(("%s depends on an unknown %s." % (desc.casual_name(), u), None))

        for err, cls in errors:
            err += " %s will not be output" % desc.casual_name()
            desc.error(err, cls=cls)

    def add_to_lookup_table(desc, kind):
        """Add `desc` to the lookup table so that other descriptions that use
        it can find it."""
        if kind == "struct":
            if (desc.variety, desc.tag) not in struct_names:
                struct_names[(desc.variety, desc.tag)] = desc
        if kind == "enum":
            if desc.tag not in enum_names:
                enum_names[desc.tag] = desc
        if kind == "typedef":
            if desc.name not in typedef_names:
                typedef_names[desc.name] = desc
        if kind in ("function", "constant", "variable", "macro"):
            if desc.name not in ident_names:
                ident_names[desc.name] = desc

    # Macros are handled differently from everything else because macros can
    # call other macros that are referenced after them in the input file, but
    # no other type of description can look ahead like that.

    for kind, desc in data.output_order:
        add_to_lookup_table(desc, kind)
        if kind != "macro":
            find_dependencies_for(desc, kind)

    for kind, desc in data.output_order:
        if kind == "macro":
            find_dependencies_for(desc, kind)



import ctypes
import sys
from ctypes import *  # noqa: F401, F403

_int_types = (ctypes.c_int16, ctypes.c_int32)
if hasattr(ctypes, "c_int64"):
    # Some builds of ctypes apparently do not have ctypes.c_int64
    # defined; it's a pretty good bet that these builds do not
    # have 64-bit pointers.
    _int_types += (ctypes.c_int64,)
for t in _int_types:
    if ctypes.sizeof(t) == ctypes.sizeof(ctypes.c_size_t):
        c_ptrdiff_t = t
del t
del _int_types


# ~POINTER~
class UserString:
    def __init__(self, seq):
        if isinstance(seq, bytes):
            self.data = seq
        elif isinstance(seq, UserString):
            self.data = seq.data[:]
        else:
            self.data = str(seq).encode()

    def __bytes__(self):
        return self.data

    def __str__(self):
        return self.data.decode()

    def __repr__(self):
        return repr(self.data)

    def __int__(self):
        return int(self.data.decode())

    def __long__(self):
        return int(self.data.decode())

    def __float__(self):
        return float(self.data.decode())

    def __complex__(self):
        return complex(self.data.decode())

    def __hash__(self):
        return hash(self.data)

    def __le__(self, string):
        if isinstance(string, UserString):
            return self.data <= string.data
        else:
            return self.data <= string

    def __lt__(self, string):
        if isinstance(string, UserString):
            return self.data < string.data
        else:
            return self.data < string

    def __ge__(self, string):
        if isinstance(string, UserString):
            return self.data >= string.data
        else:
            return self.data >= string

    def __gt__(self, string):
        if isinstance(string, UserString):
            return self.data > string.data
        else:
            return self.data > string

    def __eq__(self, string):
        if isinstance(string, UserString):
            return self.data == string.data
        else:
            return self.data == string

    def __ne__(self, string):
        if isinstance(string, UserString):
            return self.data != string.data
        else:
            return self.data != string

    def __contains__(self, char):
        return char in self.data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.__class__(self.data[index])

    def __getslice__(self, start, end):
        start = max(start, 0)
        end = max(end, 0)
        return self.__class__(self.data[start:end])

    def __add__(self, other):
        if isinstance(other, UserString):
            return self.__class__(self.data + other.data)
        elif isinstance(other, bytes):
            return self.__class__(self.data + other)
        else:
            return self.__class__(self.data + str(other).encode())

    def __radd__(self, other):
        if isinstance(other, bytes):
            return self.__class__(other + self.data)
        else:
            return self.__class__(str(other).encode() + self.data)

    def __mul__(self, n):
        return self.__class__(self.data * n)

    __rmul__ = __mul__

    def __mod__(self, args):
        return self.__class__(self.data % args)

    # the following methods are defined in alphabetical order:
    def capitalize(self):
        return self.__class__(self.data.capitalize())

    def center(self, width, *args):
        return self.__class__(self.data.center(width, *args))

    def count(self, sub, start=0, end=sys.maxsize):
        return self.data.count(sub, start, end)

    def decode(self, encoding=None, errors=None):  # XXX improve this?
        if encoding:
            if errors:
                return self.__class__(self.data.decode(encoding, errors))
            else:
                return self.__class__(self.data.decode(encoding))
        else:
            return self.__class__(self.data.decode())

    def encode(self, encoding=None, errors=None):  # XXX improve this?
        if encoding:
            if errors:
                return self.__class__(self.data.encode(encoding, errors))
            else:
                return self.__class__(self.data.encode(encoding))
        else:
            return self.__class__(self.data.encode())

    def endswith(self, suffix, start=0, end=sys.maxsize):
        return self.data.endswith(suffix, start, end)

    def expandtabs(self, tabsize=8):
        return self.__class__(self.data.expandtabs(tabsize))

    def find(self, sub, start=0, end=sys.maxsize):
        return self.data.find(sub, start, end)

    def index(self, sub, start=0, end=sys.maxsize):
        return self.data.index(sub, start, end)

    def isalpha(self):
        return self.data.isalpha()

    def isalnum(self):
        return self.data.isalnum()

    def isdecimal(self):
        return self.data.isdecimal()

    def isdigit(self):
        return self.data.isdigit()

    def islower(self):
        return self.data.islower()

    def isnumeric(self):
        return self.data.isnumeric()

    def isspace(self):
        return self.data.isspace()

    def istitle(self):
        return self.data.istitle()

    def isupper(self):
        return self.data.isupper()

    def join(self, seq):
        return self.data.join(seq)

    def ljust(self, width, *args):
        return self.__class__(self.data.ljust(width, *args))

    def lower(self):
        return self.__class__(self.data.lower())

    def lstrip(self, chars=None):
        return self.__class__(self.data.lstrip(chars))

    def partition(self, sep):
        return self.data.partition(sep)

    def replace(self, old, new, maxsplit=-1):
        return self.__class__(self.data.replace(old, new, maxsplit))

    def rfind(self, sub, start=0, end=sys.maxsize):
        return self.data.rfind(sub, start, end)

    def rindex(self, sub, start=0, end=sys.maxsize):
        return self.data.rindex(sub, start, end)

    def rjust(self, width, *args):
        return self.__class__(self.data.rjust(width, *args))

    def rpartition(self, sep):
        return self.data.rpartition(sep)

    def rstrip(self, chars=None):
        return self.__class__(self.data.rstrip(chars))

    def split(self, sep=None, maxsplit=-1):
        return self.data.split(sep, maxsplit)

    def rsplit(self, sep=None, maxsplit=-1):
        return self.data.rsplit(sep, maxsplit)

    def splitlines(self, keepends=0):
        return self.data.splitlines(keepends)

    def startswith(self, prefix, start=0, end=sys.maxsize):
        return self.data.startswith(prefix, start, end)

    def strip(self, chars=None):
        return self.__class__(self.data.strip(chars))

    def swapcase(self):
        return self.__class__(self.data.swapcase())

    def title(self):
        return self.__class__(self.data.title())

    def translate(self, *args):
        return self.__class__(self.data.translate(*args))

    def upper(self):
        return self.__class__(self.data.upper())

    def zfill(self, width):
        return self.__class__(self.data.zfill(width))


class MutableString(UserString):
    """mutable string objects

    Python strings are immutable objects.  This has the advantage, that
    strings may be used as dictionary keys.  If this property isn't needed
    and you insist on changing string values in place instead, you may cheat
    and use MutableString.

    But the purpose of this class is an educational one: to prevent
    people from inventing their own mutable string class derived
    from UserString and than forget thereby to remove (override) the
    __hash__ method inherited from UserString.  This would lead to
    errors that would be very hard to track down.

    A faster and better solution is to rewrite your program using lists."""

    def __init__(self, string=""):
        self.data = string

    def __hash__(self):
        raise TypeError("unhashable type (it is mutable)")

    def __setitem__(self, index, sub):
        if index < 0:
            index += len(self.data)
        if index < 0 or index >= len(self.data):
            raise IndexError
        self.data = self.data[:index] + sub + self.data[index + 1 :]

    def __delitem__(self, index):
        if index < 0:
            index += len(self.data)
        if index < 0 or index >= len(self.data):
            raise IndexError
        self.data = self.data[:index] + self.data[index + 1 :]

    def __setslice__(self, start, end, sub):
        start = max(start, 0)
        end = max(end, 0)
        if isinstance(sub, UserString):
            self.data = self.data[:start] + sub.data + self.data[end:]
        elif isinstance(sub, bytes):
            self.data = self.data[:start] + sub + self.data[end:]
        else:
            self.data = self.data[:start] + str(sub).encode() + self.data[end:]

    def __delslice__(self, start, end):
        start = max(start, 0)
        end = max(end, 0)
        self.data = self.data[:start] + self.data[end:]

    def immutable(self):
        return UserString(self.data)

    def __iadd__(self, other):
        if isinstance(other, UserString):
            self.data += other.data
        elif isinstance(other, bytes):
            self.data += other
        else:
            self.data += str(other).encode()
        return self

    def __imul__(self, n):
        self.data *= n
        return self


class String(MutableString, ctypes.Union):

    _fields_ = [("raw", ctypes.POINTER(ctypes.c_char)), ("data", ctypes.c_char_p)]

    def __init__(self, obj=b""):
        if isinstance(obj, (bytes, UserString)):
            self.data = bytes(obj)
        else:
            self.raw = obj

    def __len__(self):
        return self.data and len(self.data) or 0

    def from_param(cls, obj):
        # Convert None or 0
        if obj is None or obj == 0:
            return cls(ctypes.POINTER(ctypes.c_char)())

        # Convert from String
        elif isinstance(obj, String):
            return obj

        # Convert from bytes
        elif isinstance(obj, bytes):
            return cls(obj)

        # Convert from str
        elif isinstance(obj, str):
            return cls(obj.encode())

        # Convert from c_char_p
        elif isinstance(obj, ctypes.c_char_p):
            return obj

        # Convert from POINTER(ctypes.c_char)
        elif isinstance(obj, ctypes.POINTER(ctypes.c_char)):
            return obj

        # Convert from raw pointer
        elif isinstance(obj, int):
            return cls(ctypes.cast(obj, ctypes.POINTER(ctypes.c_char)))

        # Convert from ctypes.c_char array
        elif isinstance(obj, ctypes.c_char * len(obj)):
            return obj

        # Convert from object
        else:
            return String.from_param(obj._as_parameter_)

    from_param = classmethod(from_param)


def ReturnString(obj, func=None, arguments=None):
    return String.from_param(obj)


# As of ctypes 1.0, ctypes does not support custom error-checking
# functions on callbacks, nor does it support custom datatypes on
# callbacks, so we must ensure that all callbacks return
# primitive datatypes.
#
# Non-primitive return values wrapped with UNCHECKED won't be
# typechecked, and will be converted to ctypes.c_void_p.
def UNCHECKED(type):
    if hasattr(type, "_type_") and isinstance(type._type_, str) and type._type_ != "P":
        return type
    else:
        return ctypes.c_void_p


# ctypes doesn't have direct support for variadic functions, so we have to write
# our own wrapper class
class _variadic_function(object):
    def __init__(self, func, restype, argtypes, errcheck):
        self.func = func
        self.func.restype = restype
        self.argtypes = argtypes
        if errcheck:
            self.func.errcheck = errcheck

    def _as_parameter_(self):
        # So we can pass this variadic function as a function pointer
        return self.func

    def __call__(self, *args):
        fixed_args = []
        i = 0
        for argtype in self.argtypes:
            # Typecheck what we can
            fixed_args.append(argtype.from_param(args[i]))
            i += 1
        return self.func(*fixed_args + list(args[i:]))


def ord_if_char(value):
    """
    Simple helper used for casts to simple builtin types:  if the argument is a
    string type, it will be converted to it's ordinal value.

    This function will raise an exception if the argument is string with more
    than one characters.
    """
    return ord(value) if (isinstance(value, bytes) or isinstance(value, str)) else value
